@@ -9,7 +9,7 @@ use Digest::SHA 'sha256_hex','hmac_sha256','hmac_sha256_hex';
 use Date::Parse;
 use Carp 'croak';
 
-our $VERSION = '1.01';
+our $VERSION = '1.02';
 
 =head1 NAME
 
@@ -21,14 +21,14 @@ AWS::Signature4 - Create a version4 signature for Amazon Web Services
  use HTTP::Request::Common;
  use LWP;
 
- my $signer = AWS;:Signature4->new(-access_key => 'AKIDEXAMPLE',
+ my $signer = AWS::Signature4->new(-access_key => 'AKIDEXAMPLE',
                                    -secret_key => 'wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY');
- my $ua     = LW::UserAgent->new();
+ my $ua     = LWP::UserAgent->new();
 
  # Example POST request
  my $request = POST('https://iam.amazonaws.com',
 		    [Action=>'ListUsers',
-		     Version=>'2010-05-08']));
+		     Version=>'2010-05-08']);
  $signer->sign($request);
  my $response = $ua->request($request);
 
@@ -95,6 +95,7 @@ sub new {
     return bless {
 	access_key => $id,
 	secret_key => $secret,
+       (defined($args{-security_token}) ? (security_token => $args{-security_token}) : ()),
     },ref $self || $self;
 }
 
@@ -180,13 +181,32 @@ sub signed_url {
 
     return $uri if $uri->query_param('X-Amz-Signature');
 
+
+    my $scope = $self->_scope($request);
+
     $uri->query_param('X-Amz-Algorithm'  => $self->_algorithm);
-    $uri->query_param('X-Amz-Credential' => $self->access_key . '/' . $self->_scope($request));
+    $uri->query_param('X-Amz-Credential' => $self->access_key . '/' . $scope);
     $uri->query_param('X-Amz-Date'       => $self->_datetime($request));
     $uri->query_param('X-Amz-Expires'    => $expires) if $expires;
     $uri->query_param('X-Amz-SignedHeaders' => 'host');
 
-    $self->_sign($request);
+    # If there was a security token passed, we need to supply it as part of the authorization
+    # because AWS requires it to validate IAM Role temporary credentials.
+
+    if (defined($self->{security_token})) {
+        $uri->query_param('X-Amz-Security-Token' => $self->{security_token});
+    }
+
+    # Since we're providing auth via query parameters, we need to include UNSIGNED-PAYLOAD
+    # http://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-query-string-auth.html
+    # it seems to only be needed for S3.
+
+    if ($scope =~ /\/s3\/aws4_request$/) {
+        $self->_sign($request, undef, 'UNSIGNED-PAYLOAD');
+    } else {
+        $self->_sign($request);
+    }
+
     my ($algorithm,$credential,$signedheaders,$signature) =
 	$request->header('Authorization') =~ /^(\S+) Credential=(\S+), SignedHeaders=(\S+), Signature=(\S+)/;
     $uri->query_param_append('X-Amz-Signature'     => $signature);
@@ -328,13 +348,27 @@ sub _string_to_sign {
     return join("\n",'AWS4-HMAC-SHA256',$datetime,$credential_scope,$hashed_request);
 }
 
-sub _calculate_signature {
+
+=item $signing_key = AWS::Signature4->signing_key($secret_access_key,$service_name,$region,$date)
+
+Return just the signing key in the event you wish to roll your own signature.
+
+=cut
+
+sub signing_key {
     my $self = shift;
-    my ($kSecret,$service,$region,$date,$string_to_sign) = @_;
+    my ($kSecret,$service,$region,$date) = @_;
     my $kDate    = hmac_sha256($date,'AWS4'.$kSecret);
     my $kRegion  = hmac_sha256($region,$kDate);
     my $kService = hmac_sha256($service,$kRegion);
     my $kSigning = hmac_sha256('aws4_request',$kService);
+    return $kSigning;
+}
+
+sub _calculate_signature {
+    my $self = shift;
+    my ($kSecret,$service,$region,$date,$string_to_sign) = @_;
+    my $kSigning = $self->signing_key($kSecret,$service,$region,$date);
     return hmac_sha256_hex($string_to_sign,$kSigning);
 }
 
